import numpy as np
import pytest
from scipy.stats import special_ortho_group
from numpy.testing import assert_allclose

from tred import (
    generate_transform_pair_from_matrix,
    generate_dctii_m_transform_pair,
    generate_dstii_m_transform_pair,
)

GLOBAL_SEED = 1

TENSOR_SHAPES = [(4, 3, 2), (5, 7, 6), (2, 2, 6)]
MATRIX_SHAPES = [(4, 6), (7, 3)]
VECTOR_SHAPES = [(4,)]

SCIPY_GENERATORS = [
    generate_dctii_m_transform_pair,
    generate_dstii_m_transform_pair,
]


@pytest.mark.parametrize("shape", TENSOR_SHAPES + MATRIX_SHAPES + VECTOR_SHAPES)
def test_matrix_defined_transforms(shape):
    """Compare implementation with (slower) mathematically clear version"""
    # scaling constant (arbitrary)
    C = 5

    rng = np.random.default_rng(seed=GLOBAL_SEED)

    # tensors of various sizes with uniformly distributed elements within [-0.5*C, 0.5*C)
    X = rng.random(size=shape) * C - 0.5 * C
    M_mat = special_ortho_group.rvs(shape[-1], random_state=rng)

    if len(X.shape) == 3:
        n, p, t = shape
        # apply across tensor tubes
        hatX_expected = np.zeros(shape=(n, p, t))
        for i in range(n):
            for j in range(p):
                hatX_expected[i, j, :] = M_mat @ X[i, j, :]
    elif len(X.shape) == 2:
        k, t = shape
        hatX_expected = np.zeros(shape=(k, t))
        for i in range(k):
            hatX_expected[i, :] = M_mat @ X[i, :]
    elif len(X.shape) == 1:
        hatX_expected = M_mat @ X
    else:
        raise RuntimeError("Unexpected shape passed in testing module")

    # compare with our optimized implementation
    M, Minv = generate_transform_pair_from_matrix(M_mat)
    assert_allclose(hatX_expected, M(X))

    # test inverse transform working as expected
    assert_allclose(X, Minv(M(X)))


@pytest.mark.parametrize("shape", TENSOR_SHAPES + MATRIX_SHAPES)
@pytest.mark.parametrize("transform_generator", SCIPY_GENERATORS)
def test_scipy_fft_wrapper_transforms(shape, transform_generator):
    # scaling constant (arbitrary)
    C = 5

    rng = np.random.default_rng(seed=GLOBAL_SEED)

    # tensors of various sizes with uniformly distributed elements within [-0.5*C, 0.5*C)
    X = rng.random(size=shape) * C - 0.5 * C
    M, Minv = transform_generator(shape[-1])

    if len(X.shape) == 3:
        n, p, t = shape
        # apply across tensor tubes
        hatX_expected = np.zeros(shape=(n, p, t))
        for i in range(n):
            for j in range(p):
                hatX_expected[i, j:] = M(X[i, j, :])
    elif len(X.shape) == 2:
        k, t = shape
        hatX_expected = np.zeros(shape=(k, t))
        for i in range(k):
            hatX_expected[i, :] = M(X[i, :])
    else:
        raise RuntimeError("Unexpected shape passed in testing module")

    assert_allclose(hatX_expected, M(X))

    # test inverse transform working as expected
    assert_allclose(X, Minv(M(X)))



"""
The `tred` module implements tensor decomposition methods such as tensor PCA and tensor 
SVD. Most of the functionality in this module can be regarded as dimensionality reduction 
strategies for order-3 datasets of size n, p, t, where p >> n > t. 
"""

from ._m_transforms import (
    generate_transform_pair_from_matrix,
    generate_dstii_m_transform_pair,
    generate_dctii_m_transform_pair,
)
from ._tensor_ops import facewise_product, m_product
from ._tensor_pca import generate_default_m_transform_pair, tsvdm, TPCA
from ._utils import display_tensor_facewise

# private - for testing
##########################################################################################
from ._tensor_ops import _mode_1_unfold, _mode_2_unfold, _mode_3_unfold
from ._utils import _singular_vals_tensor_to_mat, _singular_vals_mat_to_tensor

__all__ = [
    "generate_transform_pair_from_matrix",
    "generate_dctii_m_transform_pair",
    "generate_dstii_m_transform_pair",
    "facewise_product",
    "m_product",
    "generate_default_m_transform_pair",
    "tsvdm",
    "TPCA",
    "display_tensor_facewise",
]



import numpy as np
import pytest

from tred import _singular_vals_tensor_to_mat, _singular_vals_mat_to_tensor

GLOBAL_SEED = 1

TENSOR_SHAPES = [(6, 4, 2), (5, 7, 6), (4, 4, 6)]


@pytest.mark.parametrize("tensor_shape", TENSOR_SHAPES)
def test_singular_value_compression(tensor_shape):
    rng = np.random.default_rng(seed=GLOBAL_SEED)
    n, p, t = tensor_shape
    k = min(n, p)

    S_mat = rng.integers(-5, 5, size=(k, t))
    S_tens = _singular_vals_mat_to_tensor(S_mat, n, p, t)

    assert np.array_equal(S_mat, _singular_vals_tensor_to_mat(S_tens))



import numpy as np
import pytest
from numpy.testing import assert_allclose

from tred import facewise_product, _mode_1_unfold, _mode_2_unfold, _mode_3_unfold

GLOBAL_SEED = 1

# various n, p, t sizes
# ensure n > p, p > n inputs are tested
TENSOR_SHAPES = [(4, 3, 2), (5, 7, 6), (2, 2, 6)]


@pytest.mark.parametrize("tensor_shape", TENSOR_SHAPES)
@pytest.mark.parametrize("include_negatives", [0, 1])
@pytest.mark.parametrize("rectangular_offset", [0, 5])
def test_facewise_product(tensor_shape, include_negatives, rectangular_offset):
    """Compare with mathematically clearer (but less efficient) implementations"""
    # scaling constants (arbitrary)
    C1 = 5
    C2 = 6

    rng = np.random.default_rng(seed=GLOBAL_SEED)
    n, p, t = tensor_shape

    # generate some compatibly sized tensors
    A = rng.random(size=(n, p, t)) * C1 - include_negatives * 0.5 * C1

    B = (
        rng.random(size=(p, n + rectangular_offset, t)) * C2
        - include_negatives * 0.5 * C2
    )

    # compute expected results using naive implementations
    fp_expected = np.zeros(shape=(n, n + rectangular_offset, t))
    for i in range(t):
        fp_expected[:, :, i] = A[:, :, i] @ B[:, :, i]

    assert_allclose(fp_expected, facewise_product(A, B))


@pytest.mark.parametrize("view", [True, False])
def test_unfolding_explicit(view):
    """Explicitly test and assert the mode-n unfolding example given in:
    Kolda, T.G. and Bader, B.W., 2009. Tensor decompositions and applications. SIAM
    review, 51(3), pp.455-500.
    """
    X = np.array(
        [
            [[1, 13], [4, 16], [7, 19], [10, 22]],
            [[2, 14], [5, 17], [8, 20], [11, 23]],
            [[3, 15], [6, 18], [9, 21], [12, 24]],
        ]
    )

    X_m1_expected = np.array(
        [
            [1, 4, 7, 10, 13, 16, 19, 22],
            [2, 5, 8, 11, 14, 17, 20, 23],
            [3, 6, 9, 12, 15, 18, 21, 24],
        ]
    )

    X_m2_expected = np.array(
        [
            [1, 2, 3, 13, 14, 15],
            [4, 5, 6, 16, 17, 18],
            [7, 8, 9, 19, 20, 21],
            [10, 11, 12, 22, 23, 24],
        ]
    )

    X_m3_expected = np.array(
        [
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24],
        ]
    )

    assert np.array_equal(_mode_1_unfold(X, view=view), X_m1_expected)
    assert np.array_equal(_mode_2_unfold(X, view=view), X_m2_expected)
    assert np.array_equal(_mode_3_unfold(X, view=view), X_m3_expected)



"""Module with mathematical operations"""


import numpy as np

from ._utils import _singular_vals_tensor_to_mat


def facewise_product(A, B):
    """Facewise product s.t. $(C_{:,:,i} = A_{:,:,i} B_{:,:,i}$.

    Parameters
    ----------
        A : ndarray, shape (a, b, d)
            Tensor represented in order-3 ndarray

        B : ndarray, shape (b, c, d)
            Tensor represented in order-3 ndarray

    Returns
    -------
        C : ndarray, shape: (a, c, d)
            Facewise tensor product
    """
    # return np.einsum('mpi,pli->mli', A, B)
    # the following is a quicker version of the above using numpy broadcasting
    return (A.transpose(2, 0, 1) @ B.transpose(2, 0, 1)).transpose(1, 2, 0)


def m_product(A, B, M, Minv):
    """Kilmer et al. (2021) tensor m-product for order-3 tensors.

    Parameters
    ----------
        A : ndarray, shape (a, b, d)
            Tensor represented in order-3 ndarray

        B : ndarray, shape (b, c, d)
            Tensor represented in order-3 ndarray

        M : Callable[[ArrayLike], ndarray]
            A function which, given some order-3 tensor, returns it under an orthogonal
            tubal transformation

        MInv : Callable[[ArrayLike], ndarray]
            A function implementing the inverse tubal transformation of M

    Returns
    -------
        m_product : ndarray, shape: (a, c, d)
            Tensor-tensor m-product as found in Kilmer et al. (2021)

    References
    ----------
    Kilmer, M.E., Horesh, L., Avron, H. and Newman, E., 2021. Tensor-tensor
    algebra for optimal representation and compression of multiway data. Proceedings
    of the National Academy of Sciences, 118(28), p.e2015851118.
    """
    assert (
        A.shape[1] == B.shape[0] and A.shape[2] == B.shape[2]
    ), "Non conforming dimensions"

    return Minv(facewise_product(M(A), M(B)))


def _rank_q_truncation_zero_out(hatU, hatS, hatV, *, q=None, sigma_q=None):
    """In-place explicit rank-q truncation as introduced in Mor et al. (2022). Truncates
    tensors U, S, V from a tsvdm decomposition to achieve an explicit rank of q.

    Parameters
    ----------
        hatU : ndarray, shape (n, k, t)
            Tensor U from the tsvdm.

        hatS : ndarray, shape (k, k, t) or (k, t)
            Tensor S from the tsvdm, or represented in compact matrix form.

        hatV : ndarray, shape (p, k, t)
            Tensor V from the tSVDM

        q : int or None, default=None
            Target explicit rank for the truncation

        sigma_q : float or None, default=None
            The `q`-th largest singular value. This will not be checked, and assumed to
            be a valid singular value in the inputted decomposition.

            If `sigma_q` is set, then the `q` input parameter will be ignored. Saves
            re-computation of the `q`-th largest singular value.

    Returns
    -------
        rho : ndarray, shape (q,)
            The multi-rank which results from the choice of `q` (or `sigma_q`)

    References
    ----------
    Mor, U., Cohen, Y., Valdés-Mas, R., Kviatcovsky, D., Elinav, E. and Avron,
    H., 2022. Dimensionality reduction of longitudinal’omics data using modern
    tensor factorizations. PLoS Computational Biology, 18(7), p.e1010212.
    """
    assert not (q is None and sigma_q is None), "Please specify either q or sigma_q"

    # assume hatS in matrix form
    if len(hatS.shape) == 3:
        hatS = _singular_vals_tensor_to_mat(hatS)

    k, t = hatS.shape

    # determine the q-th largest singular value
    if sigma_q is None:
        sigma_q = np.partition(hatS.flatten(), -q)[-q]

    # compute a representation of rho, as notated in Mor et al. (2022)
    # each diagonal of the frontal slice is already sorted in decreasing order
    # np.searchsorted is called on each column, reversed so that it is ascending,
    # where it returns the index i
    #   s.t. sigma_q <= ascending_col[i] (see numpy searchsorted docs)
    # which mean that we want to discard the i smallest values, and keep the other
    # k-i, which is the number we store in rho
    rho = np.apply_along_axis(
        lambda col: k - np.searchsorted(col[::-1], sigma_q, side="left"),
        axis=0,
        arr=hatS,
    )

    # perform multi-rank truncation, t should be modestly sized, so the for-loop
    # should be bearable
    # NOTE: there is probably some Numpy trick for this
    for i in range(t):
        hatU[:, rho[i] :, i] = 0
        hatS[rho[i] :, i] = 0
        hatV[:, rho[i] :, i] = 0

    return rho


def _mode_1_unfold(tens, view=False):
    """Return mode-1 unfolding copy, as defined in Kolda et al.

    References
    ----------
    Kolda, T.G. and Bader, B.W., 2009. Tensor decompositions and applications. SIAM
    review, 51(3), pp.455-500.
    """
    # unfold the n x p x t tensor into a n x pt 2d array (matrix), where each frontal
    # slice sits 'next' each other.
    #
    # NUMPY NOTES:
    # first transpose into a n-vertical-stack of t x p matrices. reshape is equivalent
    # to 'ravel'-ling first, before reshaping the vector (using the same memory format)
    # into the intended shape. in "C" memory format, the last indexes move the quickest,
    # so ravel moves along the p index, and then the t index, and then the n index.
    # when placing this into the matrix form, we get the intended result, where for the
    # same t and n, the p values sit next to each other; for the same n, all of its t
    # values sit together on the same row. then the matrix shape ensures that each row
    # has a distinct n
    #
    # the unfolding has the same tensor semantics as singular_values_ in the fit method
    if view:
        return tens.view().transpose(0, 2, 1).reshape((tens.shape[0], -1), order="C")
    else:
        return tens.copy().transpose(0, 2, 1).reshape((tens.shape[0], -1), order="C")


def _mode_2_unfold(tens, view=False):
    """Return mode-2 unfolding copy, as defined in Kolda et al.

    References
    ----------
    Kolda, T.G. and Bader, B.W., 2009. Tensor decompositions and applications. SIAM
    review, 51(3), pp.455-500.
    """
    if view:
        return tens.view().transpose(1, 2, 0).reshape((tens.shape[1], -1), order="C")
    else:
        return tens.copy().transpose(1, 2, 0).reshape((tens.shape[1], -1), order="C")


def _mode_3_unfold(tens, view=False):
    """Return mode-3 unfolding copy, as defined in Kolda et al.

    References
    ----------
    Kolda, T.G. and Bader, B.W., 2009. Tensor decompositions and applications. SIAM
    review, 51(3), pp.455-500.
    """
    if view:
        return tens.view().transpose(2, 1, 0).reshape((tens.shape[2], -1), order="C")
    else:
        return tens.copy().transpose(2, 1, 0).reshape((tens.shape[2], -1), order="C")



from functools import reduce

import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.stats import special_ortho_group

from tred import (
    generate_default_m_transform_pair,
    TPCA,
    tsvdm,
    m_product,
    generate_transform_pair_from_matrix,
    generate_dctii_m_transform_pair,
    generate_dstii_m_transform_pair,
)


GLOBAL_SEED = 1
RNG = np.random.default_rng(seed=GLOBAL_SEED)


def _dummy_random_orthogonal_m_transform_generator(t):
    """Let's test a m-transform defined by a random orthogonal matrix too!"""
    M_mat = special_ortho_group.rvs(t, random_state=RNG)
    M, Minv = generate_transform_pair_from_matrix(M_mat)
    return M, Minv


def _dummy_default_transform_generator(t):
    """Return M=None, Minv=None to test default m-transform configuration"""
    return None, None


# various n, p, t sizes
# ensure n > p, p > n inputs are tested
TENSOR_SHAPES = [(4, 3, 2), (5, 7, 6), (2, 2, 6)]

# m transforms to suit the tensor sizes above
TRANSFORM_FAMILY_GENERATORS = [
    _dummy_random_orthogonal_m_transform_generator,
    _dummy_default_transform_generator,
    generate_dctii_m_transform_pair,
    generate_dstii_m_transform_pair,
]

# test tiny, small, medium, and large numbers
ELEMENT_SCALES = [10**i for i in range(-4, 5, 4)]


def _check_fitted_tpca_close(tpca1, tpca2, rtol, atol):
    """Check all of the fitted attributes of the two tpca classes
    NOTE: unused at the moment, but will likely be useful in future tests
    """
    assert_allclose(tpca1.n_, tpca2.n_, rtol=rtol, atol=atol)
    assert_allclose(tpca1.p_, tpca2.p_, rtol=rtol, atol=atol)
    assert_allclose(tpca1.t_, tpca2.t_, rtol=rtol, atol=atol)
    assert_allclose(tpca1.k_, tpca2.k_, rtol=rtol, atol=atol)
    assert_allclose(tpca1.n_components_, tpca2.n_components_, rtol=rtol, atol=atol)
    assert_allclose(
        tpca1.explained_variance_ratio_,
        tpca2.explained_variance_ratio_,
        rtol=rtol,
        atol=atol,
    )
    assert_allclose(
        tpca1.singular_values_, tpca2.singular_values_, rtol=rtol, atol=atol
    )
    assert_allclose(tpca1.mean_, tpca2.mean_, rtol=rtol, atol=atol)
    assert_allclose(tpca1.rho_, tpca2.rho_, rtol=rtol, atol=atol)


@pytest.mark.parametrize("tensor_shape", TENSOR_SHAPES)
@pytest.mark.parametrize("element_scale", ELEMENT_SCALES)
@pytest.mark.parametrize("include_negatives", [0, 1])
@pytest.mark.parametrize("transform_generator", TRANSFORM_FAMILY_GENERATORS)
@pytest.mark.parametrize("return_full_frontal_slices", [True, False])
def test_tsvdm(
    tensor_shape,
    element_scale,
    include_negatives,
    transform_generator,
    return_full_frontal_slices,
):
    rng = np.random.default_rng(seed=GLOBAL_SEED)
    n, p, t = tensor_shape

    # tensors of various sizes with uniformly distributed elements
    # within [0, element_scale) or [-0.5*element_scale, 0.5*element_scale)
    X = (
        rng.random(size=tensor_shape) * element_scale
        - include_negatives * 0.5 * element_scale
    )

    M, Minv = transform_generator(t)
    U, S, V = tsvdm(X, M=M, Minv=Minv, full_frontal_slices=return_full_frontal_slices)
    Vt = V.transpose(1, 0, 2)

    # make sure that the m-product below has a 'non-None' set of inputs for M and Minv
    if M is None:
        M, Minv = generate_default_m_transform_pair(t)

    def m_product_wrapper(A, B):
        """m-product with fixed M and Minv to use for functools reduce"""
        return m_product(A, B, M=M, Minv=Minv)

    X_reconstruct = reduce(m_product_wrapper, (U, S, Vt))
    assert_allclose(X, X_reconstruct)


@pytest.mark.parametrize("tensor_shape", TENSOR_SHAPES)
@pytest.mark.parametrize("element_scale", ELEMENT_SCALES)
@pytest.mark.parametrize("include_negatives", [0, 1])
@pytest.mark.parametrize("n_components", [None, 1, 6, 0.8])
@pytest.mark.parametrize("centre", [True, False])
@pytest.mark.parametrize("transform_generator", TRANSFORM_FAMILY_GENERATORS)
def test_tpca(
    tensor_shape,
    element_scale,
    include_negatives,
    n_components,
    transform_generator,
    centre,
):
    """Make sure different inputs work, and perform basic sense checks"""
    RTOL = 1e-7
    ATOL = 1e-10

    rng = np.random.default_rng(seed=GLOBAL_SEED)

    n, p, t = tensor_shape
    k = min(n, p)

    M, Minv = transform_generator(t)
    tpca = TPCA(n_components=n_components, M=M, Minv=Minv, centre=centre)

    # tensors of various sizes with uniformly distributed elements
    # within [0, element_scale) or [-0.5*element_scale, 0.5*element_scale)
    X = (
        rng.random(size=tensor_shape) * element_scale
        - include_negatives * 0.5 * element_scale
    )

    X_r = tpca.fit(X).transform(X)

    # check the output is intended size
    assert len(X_r.shape) == 2
    assert X_r.shape[0] == n
    if n_components is None:
        assert X_r.shape[1] == k * t
    elif isinstance(n_components, int):
        assert X_r.shape[1] == n_components

    # check the equivalence of fit.transform and fit_transform
    # allow 1e-10 of absolute tolerance for small elements
    X_r2 = tpca.fit_transform(X)
    assert_allclose(X_r, X_r2, rtol=RTOL, atol=ATOL)

    # test rho
    if n_components is None:
        assert tpca.rho_.sum() == k * t
    elif isinstance(n_components, int):
        assert tpca.rho_.sum() == n_components
    else:
        # just check the sum of rho against internal n_components_
        assert tpca.rho_.sum() == tpca.n_components_

    # test explained variance ratio
    if n_components is None:
        assert_allclose(tpca.explained_variance_ratio_.sum(), 1.0)
    elif isinstance(n_components, float):
        # if n_components specifies the minimum amount of explained variance, check that
        # the truncation achieves this
        assert tpca.explained_variance_ratio_.sum() > n_components

    # test the shape of the loadings matrix
    assert tpca.loadings_matrix_.shape == (tpca.n_components_, tpca.p_)



"""Utilities for generating and defining M transforms in the m-product framework"""

import os
import sys

import numpy.linalg as linalg
from scipy.fft import dct, idct, dst, idst


# the transform generator functions below all work for:
#   1. 3D array (tensor): by applying the transform along the tubal fibres
#   2. 2D array (matrices): by applying the transform along the rows
#   3. 1D array (vector): by applying the transform on the vector
# i.e., along the -1 axis of order-2 and order-3 arrays
#
# we do not need extensive parameter validation however, as the M and Minv Callables
# returned by the generator functions are meant to be passed into classes and functions
# in the tred package - which do some of their own parameter validation
SUPPORTED_TRANSFORM_DIMENSIONS = (1, 2, 3)


def _assert_t_and_order(X_input, t_expected):
    assert (
        len(X_input.shape) in SUPPORTED_TRANSFORM_DIMENSIONS
    ), "Expecting 1D (vector), 2D (matrix) or 3D (tensor) input"
    assert (
        X_input.shape[-1] == t_expected
    ), f"Expecting last input dimension to be {t_expected}"


def generate_transform_pair_from_matrix(M_mat, Minv_mat=None, *, inplace=False):
    """Generate a pair of functions to apply a matrix, and its inverse, to the tubal
    fibres of an order-3 tensor. See Kilmer et al. (20211).

    Parameters
    ----------
        M_mat : ArrayLike
            Square matrix

        Minv_mat : ArrayLike or None, default=None
            Square matrix, the inverse of M_mat. If not specified, this function will
            numerically evaluate the inverse of `M_mat`

        inplace : bool, default=False
            Control whether or not the generated functions modify the input tensor
            in-place, or return a copy with the m-transform applied

            THIS ARGUMENT DOES NOT CURRENTLY ALTER ANY BEHAVIOUR FOR THIS FUNCTION, BUT
            IS A PLACEHOLDER FOR FUTURE DEVELOPMENT.

    Returns
    -------
        M : Callable[[ArrayLike], ndarray]
            A function which expects an order-3 tensor as input, and applies `M_mat`
            to each of the tubal fibres. This preserves the dimensions of the tensor.

        Minv : Callable[[ArrayLike], ndarray]
            A tensor transform (the inverse of `M`)

    References
    ----------
    `Kilmer, M.E., Horesh, L., Avron, H. and Newman, E., 2021. Tensor-tensor
    algebra for optimal representation and compression of multiway data. Proceedings
    of the National Academy of Sciences, 118(28), p.e2015851118.`
    """
    assert len(M_mat.shape) == 2, "Expecting matrix (order-2 array) input"
    assert M_mat.shape[0] == M_mat.shape[1], "Expecting square matrix input"
    t_ = M_mat.shape[0]

    # numerically evaluate inverse matrix if not explicitly specified
    if Minv_mat is None:
        # be pedantic and check for invertibility
        # https://stackoverflow.com/questions/13249108/efficient-pythonic-check-for-singular-matrix
        # NOTE: revisit later, is the following more robust?:
        # https://stackoverflow.com/questions/17931613/how-to-decide-a-whether-a-matrix-is-singular-in-python-numpy
        if linalg.cond(M_mat) < 1 / sys.float_info.epsilon:
            Minv_mat = linalg.inv(M_mat)
        else:
            raise ValueError(
                "Input matrix must be invertible, but appears singular, or close to singular"
            )
    else:
        assert (
            Minv_mat.shape == M_mat.shape
        ), "Ensure the shapes of matrix M and its inverse are the same"

    # a quick way of applying M along the tubal fibres of X, using np broadcasting:
    # transpose X into a n x t x p vertical stack of t x p matrices. matrix
    # multiplication is broadcast vertically, whereby M multiplies each of the t x p
    # matrices, effectively applying the transform to the columns, which were the original
    # tubal fibres. then transpose back to original
    def M(X):
        _assert_t_and_order(X, t_)
        if len(X.shape) == 3:
            return (M_mat @ X.transpose(0, 2, 1)).transpose(0, 2, 1)
        elif len(X.shape) == 2:
            return (M_mat @ X.T).T
        else:  # len(X.shape == 1)
            return M_mat @ X

    def Minv(X):
        _assert_t_and_order(X, t_)
        if len(X.shape) == 3:
            return (Minv_mat @ X.transpose(0, 2, 1)).transpose(0, 2, 1)
        elif len(X.shape) == 2:
            return (Minv_mat @ X.T).T
        else:  # len(X.shape == 1)
            return Minv_mat @ X

    return M, Minv


def generate_dctii_m_transform_pair(t, *, inplace=False, norm="ortho"):
    """Wrapper around scipy fft to generate functions to perform $\times_3 M$ operations
    on order-3 tensors, where $M$ is the (scaled) Discrete Cosine Transform.

    Parameters
    ----------
        t : int
            The length of the transform

        inplace : bool, default=False
            Control whether or not the generated functions modify the input tensor
            in-place, or return a copy with the m-transform applied

        norm : {“backward”, “ortho”, “forward”}, default="ortho"
            See https://docs.scipy.org/doc/scipy/reference/generated/scipy.fft.dct.html#scipy.fft.dct

    Returns
    -------
        M : Callable[[ArrayLike], ndarray]
            A function which expects an order-3 tensor as input, and applies DCT-II to
            each of the tubal fibres. This preserves the dimensions of the tensor.

        Minv : Callable[[ArrayLike], ndarray]
            A tensor transform (the inverse of `M`)

    References
    ----------
    The use of this transform with the m-product was introduced in:
    `Mor, U., Cohen, Y., Valdés-Mas, R., Kviatcovsky, D., Elinav, E. and Avron,
    H., 2022. Dimensionality reduction of longitudinal’omics data using modern
    tensor factorizations. PLoS Computational Biology, 18(7), p.e1010212.`
    """

    def M(X):
        _assert_t_and_order(X, t)
        return dct(
            X,
            type=2,
            n=t,
            axis=-1,
            norm=norm,
            workers=2 * os.cpu_count(),
            overwrite_x=inplace,
        )

    def Minv(X):
        _assert_t_and_order(X, t)
        return idct(
            X,
            type=2,
            n=t,
            axis=-1,
            norm=norm,
            workers=2 * os.cpu_count(),
            overwrite_x=inplace,
        )

    return M, Minv


def generate_dstii_m_transform_pair(t, *, inplace=False, norm="ortho"):
    """Wrapper around scipy fft to generate functions to perform $\times_r M$ operations
    on order-3 tensor, where $M$ is the (scaled) Discrete Sine Transform.

    Parameters
    ----------
        t : int
            The length of the transform

        inplace : bool, default=False
            Control whether or not the generated functions modify the input tensor
            in-place, or return a copy with the m-transform applied

        norm : {“backward”, “ortho”, “forward”}, default="ortho"
            See https://docs.scipy.org/doc/scipy/reference/generated/scipy.fft.dst.html#scipy.fft.dst

    Returns
    -------
        M : Callable[[ArrayLike], ndarray]
            A function which expects an order-3 tensor as input, and applies DFT via a FFT
            algorithm to each of the tubal fibres. This preserves the dimensions of the
            tensor.

        Minv : Callable[[ArrayLike], ndarray]
            A tensor transform (the inverse of `M`)
    """

    def M(X):
        _assert_t_and_order(X, t)
        return dst(
            X,
            type=2,
            n=t,
            axis=-1,
            norm=norm,
            workers=2 * os.cpu_count(),
            overwrite_x=inplace,
        )

    def Minv(X):
        _assert_t_and_order(X, t)
        return idst(
            X,
            type=2,
            n=t,
            axis=-1,
            norm=norm,
            workers=2 * os.cpu_count(),
            overwrite_x=inplace,
        )

    return M, Minv
